﻿//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Anwendung                                                                           //
//  _________________________________                                                   //
//                                                                                      //
//  Copyright (c) 2019 Reimar Grasbon                                                   //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  This software is subjected to the FreeBSD License. Redistribution and use in source //
//  and binary forms, with or without modification, are permitted provided that the     //
//  following conditions are met:                                                       //
//                                                                                      //
//  1. Redistributions of source code must retain the above copyright notice, this list //
//     of conditions and the following disclaimer.                                      //
//                                                                                      //
//  2. Redistributions in binary form must reproduce the above copyright notice, this   //
//     list of conditions and the following disclaimer in the documentation and/or      //
//     other materials provided with the distribution.                                  //
//                                                                                      //
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY //
//  EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED           //
//  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  //
//  IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,    //
//  INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT  //
//  NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR  //
//  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,   //
//  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)  //
//  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE          //
//  POSSIBILITY OF SUCH DAMAGE.                                                         //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////

#include <avr/pgmspace.h>
#include "Cy4Net.h"
#include "Cy4TypePlate.h"
#include "Storing.h"
#include "Board.h"
#include "Timer.h"
#include "Application.h"

//////////////////////////////////////////////////////////////////////////////////////////
//  Typen  ///////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
typedef enum                        // Kontaktereignis:
{   CONTACTEVENT_CLOSED,            // · Kontakt wurde geschlossen
    CONTACTEVENT_OPENED,            // · Kontakt wurde geöffnet
    CONTACTEVENT_NONE               // · Kein Ereignis
} ContactEvent;

typedef enum                        // Kontaktzustand:
{   CONTACTSTATE_CLOSED,            // · Kontakt ist geschlossen
    CONTACTSTATE_OPENED,            // · Kontakt ist geöffnet
    CONTACTSTATE_UNKNOWN            // · Kontaktzustand unbekannt (Anfangszustand)
} ContactState;

//////////////////////////////////////////////////////////////////////////////////////////
//  Daten  ///////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
static ContactEvent        contactEvent[CFG_CONTACT_COUNT]       = { CONTACTEVENT_NONE,
                                                                     CONTACTEVENT_NONE,
                                                                     CONTACTEVENT_NONE
                                                                   };
static ContactState        contactState[CFG_CONTACT_COUNT]       = { CONTACTSTATE_UNKNOWN,
                                                                     CONTACTSTATE_UNKNOWN,
                                                                     CONTACTSTATE_UNKNOWN
                                                                   };
static const PushMessage*  currentPushMessage[CFG_CONTACT_COUNT] = { NULL,
                                                                     NULL,
                                                                     NULL
                                                                   };
static const TextFeed*     currentTextFeed[CFG_CONTACT_COUNT]    = { NULL,
                                                                     NULL,
                                                                     NULL
                                                                   };

//////////////////////////////////////////////////////////////////////////////////////////
//  Cy4-Nachrichten-Ausstatter  //////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
static void
SetupContact1PushMessage(Cy4Message* message)
{
    PushMessage pushMessage;
    PUSHMESSAGE_EEPLoad(currentPushMessage[0], &pushMessage);
    PUSHMESSAGE_ToMessage(&pushMessage, message);
}

//////////////////////////////////////////////////////////////////////////////////////////
static void
SetupContact2PushMessage(Cy4Message* message)
{
    PushMessage pushMessage;
    PUSHMESSAGE_EEPLoad(currentPushMessage[1], &pushMessage);
    PUSHMESSAGE_ToMessage(&pushMessage, message);
}

//////////////////////////////////////////////////////////////////////////////////////////
static void
SetupContact3PushMessage(Cy4Message* message)
{
    PushMessage pushMessage;
    PUSHMESSAGE_EEPLoad(currentPushMessage[2], &pushMessage);
    PUSHMESSAGE_ToMessage(&pushMessage, message);
}

//////////////////////////////////////////////////////////////////////////////////////////
static void
SetupContact1TextFeed(Cy4Message* message)
{
    TextFeed textFeed;
    TEXTFEED_EEPLoad(currentTextFeed[0], &textFeed);
    TEXTFEED_ToMessage(&textFeed, message);
}

//////////////////////////////////////////////////////////////////////////////////////////
static void
SetupContact2TextFeed(Cy4Message* message)
{
    TextFeed textFeed;
    TEXTFEED_EEPLoad(currentTextFeed[1], &textFeed);
    TEXTFEED_ToMessage(&textFeed, message);
}

//////////////////////////////////////////////////////////////////////////////////////////
static void
SetupContact3TextFeed(Cy4Message* message)
{
    TextFeed textFeed;
    TEXTFEED_EEPLoad(currentTextFeed[2], &textFeed);
    TEXTFEED_ToMessage(&textFeed, message);
}

//////////////////////////////////////////////////////////////////////////////////////////
//  Verfahrensfunktionen  ////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////////////////////////
//  ProceedContacting(): Verfahrensfunktion der Kontaktierung. Die Zustand des Auto-    //
//  -------------------  maten Automaten findet sich in <contactState>.                 //
//                                                                                      //
//   Zustandsübergangsdiagramm:                                                         //
//                                                   PinState=false,                    //
//                                      +----------+ StartTimer()                       //
//                  +------------------>|  OPENED  +-----------------+                  //
//                  | State: Opened     +----------+                 |                  //
//                  | Event: Opened           /|\                    |                  //
//                  | TIMEOUT                  |                    \|/                 //
//            +-----+-----+                    |  PinState=true +---------+             //
//            |  OPENING  +-----------------+  +----------------+ CLOSING |             //
//            +-----------+ PinState=false  |                   +----+----+             //
//                 /|\                      |                        | TIMEOUT          //
//                  |                      \|/                       | State: Closed    //
//                  |                   +----------+                 | Event: Closed    //
//                  +-------------------|  CLOSED  |<----------------+                  //
//                      PinState=true,  +----------+                                    //
//                      StartTimer()                                                    //
//                                                                                      //
//  Eingabe: -                                                                          //
//  Ausgabe: -                                                                          //
//////////////////////////////////////////////////////////////////////////////////////////
static void
ProceedContacting(void)
{
    typedef enum     // Kontaktierungs-Zustände:
    {   OPENED,      // · Kontakt ist geöffnet
        CLOSING,     // · Kontakt schliessend
        CLOSED,      // · Kontakt ist geschlossen
        OPENING,     // · Kontakt öffnend
        UNKNOWN      // · Kontaktzustand unbekannt (Anfangszustand)
    } State;

    static Timer durationTimer[CFG_CONTACT_COUNT] = { TIMER_SECONDS(0),
                                                      TIMER_SECONDS(0),
                                                      TIMER_SECONDS(0)
                                                    };
    static State state[CFG_CONTACT_COUNT]         = { UNKNOWN, UNKNOWN, UNKNOWN };
    for (UInt08 i=0; i<CFG_CONTACT_COUNT; i++)
    {   contactEvent[i] = CONTACTEVENT_NONE;     // Ereignis existiert nur einen Durchlauf
        bool isOpen = BOARD_PinState(i);
        switch (state[i])
        {   case OPENED:
                if( !isOpen )
                {   UInt08 duration;
                    STORING_Load(STO_ContactEventConfiguration[i][0].duration, &duration);
                    TIMER_Start(&durationTimer[i], duration);
                    state[i] = CLOSING;
                }
                break;
            case CLOSING:
                if( isOpen )
                {   state[i] = OPENED;
                }
                else if( TIMER_HasExpired(&durationTimer[i]) )
                {   state[i] = CLOSED;
                    contactState[i] = CONTACTSTATE_CLOSED;
                    contactEvent[i] = CONTACTEVENT_CLOSED;
                }
                break;
            case CLOSED:
                if( isOpen )
                {   UInt08 duration;
                    STORING_Load(STO_ContactEventConfiguration[i][1].duration, &duration);
                    TIMER_Start(&durationTimer[i], duration);
                    state[i] = OPENING;
                }
                break;
            case OPENING:
                if( !isOpen )
                {   state[i] = CLOSED;
                }
                else if( TIMER_HasExpired(&durationTimer[i]) )
                {   state[i] = OPENED;
                    contactState[i] = CONTACTSTATE_OPENED;
                    contactEvent[i] = CONTACTEVENT_OPENED;
                }
                break;
            default:
                // UNKNOWN
                if( isOpen )
                {   state[i] = OPENED;
                    contactState[i] = CONTACTSTATE_OPENED;
                }
                else
                {   state[i] = CLOSED;
                    contactState[i] = CONTACTSTATE_CLOSED;
                }
        }
    }
}

//////////////////////////////////////////////////////////////////////////////////////////
static void
ProceedSignaling(void)
{	
    // Ereignis-Signalausgabe
    for (UInt08 i=0; i<CFG_CONTACT_COUNT; i++)
    {   UInt08 index = 255;
        switch (contactEvent[i])
        {   case CONTACTEVENT_CLOSED: index = 0; break;
            case CONTACTEVENT_OPENED: index = 1; break;
            default:                  ;
        }
        if (index != 255)
        {   ToneEventSignal toneSignal;
            STORING_Load(STO_ContactEventConfiguration[i][index].toneSignal, &toneSignal);
            TONESIGNAL_Play(toneSignal);
            LightEventSignal lightSignal;
            STORING_Load(STO_ContactEventConfiguration[i][index].lightSignal, &lightSignal);
			UInt08 brightness;
			STORING_Load(STO_ContactEventConfiguration[i][index].brightness, &brightness);
            LIGHTSIGNAL_Play(lightSignal, brightness);
        }
    }
    // Zustands-Signalausgabe
    {	ToneStateSignal  toneSignal  = TONESTATESIGNAL_OFF;
		LightStateSignal lightSignal = LIGHTSTATESIGNAL_OFF;
		UInt08           brightness  = 10;
		for (UInt08 i=0; i<CFG_CONTACT_COUNT; i++)
		{   UInt08 index = 255;
			switch (contactState[i])
			{   case  CONTACTSTATE_CLOSED: index = 0; break;
				case  CONTACTSTATE_OPENED: index = 1; break;
				default:                   ;
			}
			if (index != 255)
			{   ContactStateSignal tmpSignal;
				STORING_Load(STO_ContactStateSignal[i][index], &tmpSignal);
				if (toneSignal  == TONESTATESIGNAL_OFF)  toneSignal  = tmpSignal.toneSignal;
				if (lightSignal == LIGHTSTATESIGNAL_OFF)
				{	lightSignal = tmpSignal.lightSignal;
					brightness = tmpSignal.brightness;
				}
			}
		}
		if (TONESIGNAL_Current() != toneSignal)
		{	TONESIGNAL_Start(toneSignal);
		}
		if (LIGHTSIGNAL_Current() != lightSignal || LIGHTSIGNAL_Brightness() != brightness)
		{	LIGHTSIGNAL_Start(lightSignal, brightness);
		}
	}
}

//////////////////////////////////////////////////////////////////////////////////////////
static void
ProceedPushMessaging(void)
{
    for (UInt08 i=0; i<CFG_CONTACT_COUNT; i++)
    {   const PushMessage* eepPushMessage = NULL;
        switch (contactEvent[i])
        {   case CONTACTEVENT_CLOSED:
                eepPushMessage = STORING_AddressOf(STO_ContactEventConfiguration[i][0].pushMessage);
                break;
            case CONTACTEVENT_OPENED:
                eepPushMessage = STORING_AddressOf(STO_ContactEventConfiguration[i][1].pushMessage);
                break;
            default:
                // CONTACTEVENT_NONE
                ;
        }
        if (eepPushMessage != NULL)
        {   PushMessage pushMessage;
            PUSHMESSAGE_EEPLoad(eepPushMessage, &pushMessage);
            if (PUSHMESSAGE_IsValid(&pushMessage))
            {   currentPushMessage[i] = eepPushMessage;
                static const Cy4MessageSetup  messageSetupTable[CFG_CONTACT_COUNT] PROGMEM =
                {   SetupContact1PushMessage,
                    SetupContact2PushMessage,
                    SetupContact3PushMessage
                };
                Cy4MessageSetup messageSetup = (Cy4MessageSetup)pgm_read_word(&messageSetupTable[i]);
                if (!CY4NET_IsPending(messageSetup, NULL)) CY4NET_Post(messageSetup, NULL);
            }
        }
    }
}

//////////////////////////////////////////////////////////////////////////////////////////
static void
ProceedTextFeeds(void)
{
    for (UInt08 i=0; i<CFG_CONTACT_COUNT; i++)
    {   const TextFeed* eepTextFeed = NULL;
        switch (contactEvent[i])
        {   case CONTACTEVENT_CLOSED:
                eepTextFeed = STORING_AddressOf(STO_ContactEventConfiguration[i][0].textFeed);
                break;
            case CONTACTEVENT_OPENED:
                eepTextFeed = STORING_AddressOf(STO_ContactEventConfiguration[i][1].textFeed);
                break;
            default:
                // CONTACTEVENT_NONE
                ;
        }
        if (eepTextFeed != NULL)
        {   TextFeed textFeed;
            TEXTFEED_EEPLoad(eepTextFeed, &textFeed);
            if (!TEXTFEED_IsEmpty(&textFeed))
            {   currentTextFeed[i] = eepTextFeed;
                static const Cy4MessageSetup  messageSetupTable[CFG_CONTACT_COUNT] PROGMEM =
                {   SetupContact1TextFeed,
                    SetupContact2TextFeed,
                    SetupContact3TextFeed
                };
                Cy4MessageSetup messageSetup = (Cy4MessageSetup)pgm_read_word(&messageSetupTable[i]);
                if (!CY4NET_IsPending(messageSetup, NULL)) CY4NET_Post(messageSetup, NULL);
            }
        }
    }
}

//////////////////////////////////////////////////////////////////////////////////////////
//  Anfrage-Bearbeiter  //////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Dienst: ContactState (read only)                                                    //
//                                                                                      //
//  Beschreibung: Ermittelt den Zustand der Kontakte unter Brücksichtigung der Er-      //
//  eignis-Verzögerungen (s. ContactEventConfiguration).                                //
//                                                                                      //
//              0   1   2                                                               //
//            +---+---+---+                                                             //
//  Belegung: |C1S|C2S|C3S|                                                             //
//            +---+---+---+                                                             //
//                                                                                      //
//  [CnS] Zustand Kontakt n: 0|1 = geöffnet|geschlossen.                                //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
static void
OnContactState(const Cy4Service* service, Cy4Message* message)
{
    UInt08  data[CFG_CONTACT_COUNT];

    for (UInt08 i=0; i<CFG_CONTACT_COUNT; i++)
    {   bool closed;
        switch (contactState[i])
        {   case CONTACTSTATE_OPENED:  closed = false; break;
            case CONTACTSTATE_CLOSED:  closed = true;  break;
            default:                   closed = false;
        }
        data[i] = closed;
    }
    CY4SERVICE_TransferPayload(service, message, data);
}

//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Dienst: ContactStateSignal (read write)                                             //
//                                                                                      //
//  Beschreibung: Legt die Art des Licht- und Tonsignals fest, das bei Zustand des      //
//                entsprechenden Kontaktes ausgegeben wird. Der Kontaktzustand wird     //
//                unter Berücksichtigung der Eregniskonfiguration ermittelt (siehe      //
//                Dienst ContactEventConfiguration).                                    //
//                                                                                      //
//              0   1   2   3   4   5                                                   //
//            +---+---+---+---+---+---+  6   7   8   9   10  11                         //
//  Belegung: |T1C|L1C|B1C|T1O|L1O|B1O|+---+---+---+---+---+---+                        //
//            +---+---+---+---+---+---+|T2C|L2C|B2C|T2O|L2O|B2O| ...                    //
//                                     +---+---+---+---+---+---+                        //
//                                                  									//
//                                                         12  13  14  15  16  17		//
//                                                       +---+---+---+---+---+---+		//
//                                                  ...  |T3C|L3C|B3C|T3O|L3O|B3O|		//
//                                                       +---+---+---+---+---+---+		//
//                                                                                      //
//  [TnC] Tonsignal, wenn Kontakt n geschlossen ist.                                    //
//          0 = Aus (Voreinstellung)                                                    //
//          1 = Tropfen                                                                 //
//          2 = Ticken                                                                  //
//          3 = Martinticks                                                             //
//          4 = Kurzer Piepton D7                                                       //
//          5 = Langer Piepton D7                                                       //
//          * = Aus                                                                     //
//  [LnC] Lichtsignal, wenn Kontakt n geschlossen ist.                                  //
//          0 = Kein Signal, LED aus (Voreinstellung).                                  //
//          1 = Ein                                                                     //
//          2 = Blinkt                                                                  //
//          3 = Blitzt                                                                  //
//          4 = Pulsiert                                                                //
//          5 = Herzschlag                                                              //
//          * = Kein Signal, LED aus.                                                   //
//  [BnC] Helligkeit der LED bei Lichtsignal [LnC]: 10 .. 100 %					        //
//  [TnO] Tonsignal, wenn Kontakt n geöffnet ist (Wert wie bei TnC)                     //
//  [LnO] Lichtsignal, wenn Kontakt n geöffnet ist (Wert wie bei LnC).                  //
//  [BnO] Helligkeit der LED bei Lichtsignal [LnO]: 10 .. 100 %					        //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
static void
OnContactStateSignal(const Cy4Service* service, Cy4Message* message)
{
    UInt08  data[3*6];
    for (UInt08 i=0; i<CFG_CONTACT_COUNT; i++)
    {   ContactStateSignal signal;
        // Geschlossen
        STORING_Load(STO_ContactStateSignal[i][0], &signal);
        data[6*i+0] = (UInt08)signal.toneSignal;
        data[6*i+1] = (UInt08)signal.lightSignal;
		data[6*i+2] = (UInt08)signal.brightness;
        // Geöffnet
        STORING_Load(STO_ContactStateSignal[i][1], &signal);
        data[6*i+3] = (UInt08)signal.toneSignal;
        data[6*i+4] = (UInt08)signal.lightSignal;
		data[6*i+5] = (UInt08)signal.brightness;
    }
    CY4SERVICE_TransferPayload(service, message, data);
    for (UInt08 i=0; i<CFG_CONTACT_COUNT; i++)
    {   ContactStateSignal signal;
        // Geschlossen
        if (data[6*i+0] < (UInt08)TONESTATESIGNAL_NUMBER)  signal.toneSignal  = (ToneStateSignal)data[6*i+0];
        else                                               signal.toneSignal  = TONESTATESIGNAL_OFF;
        if (data[6*i+1] < (UInt08)LIGHTSTATESIGNAL_NUMBER) signal.lightSignal = (LightStateSignal)data[6*i+1];
        else                                               signal.lightSignal = LIGHTSTATESIGNAL_OFF;		
		if      (data[6*i+2] <  10) signal.brightness =  10;
		else if (data[6*i+2] > 100) signal.brightness = 100;
		else                        signal.brightness = (UInt08)data[6*i+2];		
        STORING_Save(&signal, STO_ContactStateSignal[i][0]);
        // Geöffnet
        if (data[6*i+3] < (UInt08)TONESTATESIGNAL_NUMBER)  signal.toneSignal  = (ToneStateSignal)data[6*i+3];
        else                                               signal.toneSignal  = TONESTATESIGNAL_OFF;
        if (data[6*i+4] < (UInt08)LIGHTSTATESIGNAL_NUMBER) signal.lightSignal = (LightStateSignal)data[6*i+4];
        else                                               signal.lightSignal = LIGHTSTATESIGNAL_OFF;
		if      (data[6*i+5] <  10) signal.brightness =  10;
		else if (data[6*i+5] > 100) signal.brightness = 100;
		else                        signal.brightness = (UInt08)data[6*i+5];
        STORING_Save(&signal, STO_ContactStateSignal[i][1]);
    }
}

//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Dienst: ContactEventConfiguration (read write)                                      //
//                                                                                      //
//  Beschreibung: Bestimmt Verhalten und Charakteristik des jeweiligen Kontakt-         //
//                wechsels. ContactEventConfiguration ist ein Gemeinschaftsbearbeiter   //
//                für folgende Dienste:                                                 //
//                                                                                      //
//                  130100h:61: Parameter bei Ereignis Kontakt 1 geschlossen            //
//                  130180h:61: Parameter bei Ereignis Kontakt 1 geöffnet               //
//                  130200h:61: Parameter bei Ereignis Kontakt 2 geschlossen            //
//                  130280h:61: Parameter bei Ereignis Kontakt 2 geöffnet               //
//                  130300h:61: Parameter bei Ereignis Kontakt 3 geschlossen            //
//                  130380h:61: Parameter bei Ereignis Kontakt 3 geöffnet               //
//                                                                                      //
//              0         2   3                             10  11  12        59  60    //
//            +---+  1  +---+---+  4   5   6   7   8   9  +---+---+---+-   -+---+---+   //
//  Belegung: |DUR|+---+|LSE|BRT|+---+---+---+---+---+---+|T00|T01|T02| ... |T49| 0 |   //
//            +---+|TSE|+---+---+|NET|DEV|LC0|LC1|LC2|PAR|+---+---+---+-   -+---+---+   //
//                 +---+         +---+---+---+---+---+---+                              //
//                                                                                      //
//  [DUR] Zustandsdauer. Bestimmt die Dauer, wie lange der jeweilige Kontaktzustand     //
//        anhalten muss, bis das entsprechende Ereignis eintritt: 0 .. 240 [s].         //
//  [TSE] Akustisches Signal des Piezosignalgebers bei o.g. Ereignis:                   //
//           0 = Kein Signal (Voreinstellung).                                          //
//           1 = Tief Hoch                                                              //
//           2 = Hoch Tief                                                              //
//           3 = Tief Mitte Hoch                                                        //
//           4 = Hoch Mitte Tief                                                        //
//           5 = Taste gedrückt                                                         //
//           6 = Oktave2Hoch                                                            //
//           7 = Doppelton                                                              //
//           8 = Tripelton                                                              //
//           9 = Zwitscher                                                              //
//          10 = Kurzer Piepton C8                                                      //
//          11 = Langer Piepton C8                                                      //
//          12 = Kurzer Piepton C6                                                      //
//          13 = Langer Piepton C6                                                      //
//           * = Kein Signal, Stille                                                    //
//  [LSE] Lichtsignaleffekt der LED bei o.g. Ereignis:                                  //
//           0 = LED aus (Voreinstellung).                                              //
//           1 = Blinkt einmal                                                          //
//           2 = Blinkt zweimal                                                         //
//           3 = Blinkt dreimal                                                         //
//           4 = Blinkt sechsmal                                                        //
//           5 = Blitzt einmal                                                          //
//           6 = Blitzt zweimal                                                         //
//           7 = Blitzt dreimal                                                         //
//           8 = Blitzt sechsmal                                                        //
//           9 = Einblenden                                                             //
//          10 = Ausblenden                                                             //
//          11 = Flackern                                                               //
//           * = Kein Signal, LED aus.                                                  //
//  [BRT] Helligkeit der LED beim Lichtsignaleffekt [LSE]: 10 .. 100 %					//
//  [NET] Push-Nachricht Ziel bei o.g. Ereignis, Netzadresse   : 0 ..     255           //
//  [DEV] Push-Nachricht Ziel bei o.g. Ereignis, Geräteadresse : 0 ..     255           //
//  [LCx] Push-Nachricht Ziel bei o.g. Ereignis, Dienstort     : 0 .. 0xFFFFF           //
//  [PAR] Push-Nachricht Ziel bei o.g. Ereignis, Parameter-Byte: 0 ..     255           //
//  [Txx] Zeichen der Rundmitteilung bei o.g. Ereignis: 0 .. 255                        //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
static void
OnContactEventConfiguration(const Cy4Service* service, Cy4Message* message)
{
    UInt32 location = CY4MESSAGE_Location(message);
    void* eepAddress;
    if      (location < 0x130180) eepAddress = STORING_AddressOf(STO_ContactEventConfiguration[0][0]);
    else if (location < 0x130200) eepAddress = STORING_AddressOf(STO_ContactEventConfiguration[0][1]);
    else if (location < 0x130280) eepAddress = STORING_AddressOf(STO_ContactEventConfiguration[1][0]);
    else if (location < 0x130300) eepAddress = STORING_AddressOf(STO_ContactEventConfiguration[1][1]);
    else if (location < 0x130380) eepAddress = STORING_AddressOf(STO_ContactEventConfiguration[2][0]);
    else                          eepAddress = STORING_AddressOf(STO_ContactEventConfiguration[2][1]);
    ContactEventConfiguration configuration;
    STORING_LoadEEPMem(eepAddress, &configuration, sizeof(ContactEventConfiguration));
    UInt08  data[61];
    data[0] = configuration.duration;
    data[1] = (UInt08)configuration.toneSignal;
    data[2] = (UInt08)configuration.lightSignal;
	data[3] = (UInt08)configuration.brightness;
    PUSHMESSAGE_ToData(&configuration.pushMessage, &data[4]);
    TEXTFEED_ToData(&configuration.textFeed, &data[10]);
    CY4SERVICE_TransferPayload(service, message, data);
    if (data[0] > 240) configuration.duration = 240;
    else               configuration.duration = data[0];
    if (data[1] < (UInt08)TONEEVENTSIGNAL_NUMBER) configuration.toneSignal = (ToneEventSignal)data[1];
    else                                          configuration.toneSignal = TONEEVENTSIGNAL_NONE;
    if (data[2] < (UInt08)LIGHTEVENTSIGNAL_NUMBER) configuration.lightSignal = (LightEventSignal)data[2];
    else                                           configuration.lightSignal = LIGHTEVENTSIGNAL_NONE;
	if      (data[3] <  10) configuration.brightness =  10;
	else if (data[3] > 100) configuration.brightness = 100;
	else                    configuration.brightness = (UInt08)data[3];	
    PUSHMESSAGE_FromData(&data[4], &configuration.pushMessage);
    TEXTFEED_FromData(&data[10], &configuration.textFeed);
    STORING_SaveEEPMem(&configuration, eepAddress, sizeof(ContactEventConfiguration));
}

//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Dienst: ContactEventTest (write only)                                               //
//                                                                                      //
//  Beschreibung: Löst zu Probewecken die eingestellte Kontakt-Ereignis-Signalisierung	//
//				  aus.																    //
//                                                                                      //
//              0   1             4   5                                                 //
//            +---+---+  2   3  +---+---+                                               //
//  Belegung: |CE1|OE1|+---+---+|CE3|OE3|                                               //
//            +---+---+|CE2|OE2|+---+---+                                               //
//					   +---+---+														//
//                                                                                      //
//  [CEn] Auslöser der Signalisierung Ereignis Kontakt n geschlossen: 94 (5Eh, '^').    //
//  [OEn] Auslöser der Signalisierung Ereignis Kontakt n geöffnet: 94 (5Eh, '^').       //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
static void
OnContactEventTest(const Cy4Service* service, Cy4Message* message)
{
	UInt08  data[] = { 0, 0, 0, 0, 0, 0 };
	CY4SERVICE_TransferPayload(service, message, data);
	UInt08 i = CARDINALITY(data);
	do
	{	i--;
		if (data[i] == 0x5E)
		{	UInt08 cc = i/2;	// Kontaktnummer: 0 .. 2
			UInt08 ce = i&1;	// Kontakt-Ereignis: 0 = geschlossen, 1 = geöffnet
			ContactEventConfiguration configuration;			
			STORING_Load(STO_ContactEventConfiguration[cc][ce], &configuration);
			LIGHTSIGNAL_Play(configuration.lightSignal, configuration.brightness);
			TONESIGNAL_Play(configuration.toneSignal);
			break; // Auslösen mehrerer Tests ergibt keinen Sinn: Bearbeiter verlassen
		}
	} while (i != 0);
}

//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Dienst: LedBusActivity (read write)                                                 //
//                                                                                      //
//  Beschreibung: Bestimmt, ob und wie hell Bus-Aktivität mit dem Gerät wird auf der	//
//                LED angezeigt wird.                                                   //
//                                                                                      //
//              0   1                                                                   //
//            +---+---+                                                                 //
//  Belegung: |ACT|BRT|                                                                 //
//            +---+---+                                                                 //
//                                                                                      //
//  [ACT] Bus-Aktivität wird auf der LED angezeigt: 0 = Nein, sonst Ja.					//
//  [BRT] Helligkeit der LED bei Bus-Aktivität: 10 .. 100 %			                    //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
static void
OnBusActivityOnLed(const Cy4Service* service, Cy4Message* message)
{    	
	LedBusActivity ledBusActivity;
    STORING_Load(STO_LedBusActivity, &ledBusActivity);
	UInt08 data[2];
	data[0] = (bool)ledBusActivity.enabled;
	data[1] = ledBusActivity.brightness;
    CY4SERVICE_TransferPayload(service, message, data);
	ledBusActivity.enabled = (bool)data[0];	
	ledBusActivity.brightness = data[1];
	if      (data[1] <  10) ledBusActivity.brightness =  10;
	else if (data[1] > 100) ledBusActivity.brightness = 100;
	else                    ledBusActivity.brightness = data[1];  
	STORING_Save(&ledBusActivity, STO_LedBusActivity);
}

//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Dienst: DeviceLabel (read write)                                                    //
//                                                                                      //
//  Beschreibung: Benutzerdefinierte Gerätebezeichnung.                                 //
//                                                                                      //
//              0   1   2   ...  59  60                                                 //
//            +---+---+---+-   -+---+---+                                               //
//  Belegung: |N01|N02|N03| ... |N60| 0 |                                               //
//            +---+---+---+-   -+---+---+                                               //
//                                                                                      //
//  [Nxx] Zeichen der allgemeinen Gerätebezeichnung: 0 .. 255 (ASCII/UTF-8)             //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
static void
OnDeviceLabel(const Cy4Service* service, Cy4Message* message)
{
    UInt08  label[61];

    STORING_Load(STO_DeviceLabel, label);
    CY4SERVICE_TransferPayload(service, message, label);
    label[sizeof(label)-1] = '\0';
    STORING_Save(label, STO_DeviceLabel);
}

//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Dienst: FactorySettings (write only)                                                //
//                                                                                      //
//  Beschreibung: Zurückstellen aller Einstellungen auf den Ursprungszustand.           //
//                                                                                      //
//              0                                                                       //
//            +---+                                                                     //
//  Belegung: |FST|                                                                     //
//            +---+                                                                     //
//                                                                                      //
//  [FST] Auslöser zur Wiederherstellung der Werksvoreinstellungen: 94 (5Eh, '^').      //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
static void
OnFactorySettings(const Cy4Service* service, Cy4Message* message)
{
    (void)service;
    if (CY4MESSAGE_ConstData(message)[0] == 0x5E) STORING_RestoreDefaults();
    else                                          CY4MESSAGE_Invalidate(message);
}

//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Pflichtdienst: CallTypePlate (read only)                                            //
//                                                                                      //
//  Beschreibung: Geräte-Typenschild erfragen.                                          //
//                                                                                      //
//              0   1   2   3   4   5   6                                               //
//            +---+---+---+---+---+---+---+                                             //
//  Belegung: |AT2|AT1|AT0|AVE|BT2|BT1|BT0|                                             //
//            +---+---+---+---+---+---+---+                                             //
//                                                                                      //
//  [ATx] Anwendungstypennummer: Axxxxxh, xxxxx in BCD: 00001 .. 99999                  //
//  [AVE] Anwendungsversion:     01 .. 99, Haupt/Unternummer in BCD (Bsp: 12h = V1.2)   //
//  [BTx] Boardtypennummer:      Bxxxxxh, xxxxx in BCD: 00001 .. 99999                  //
//                                                                                      //
//                                                                                      //
//      Beispiel:                                                                       //
//                                                                                      //
//          +---+---+---+---+---+---+---+                                               //
//          |A0h|01h|23h|45h|B0h|67h|89h|      Anwendungstypennummer: 123               //
//          +---+---+---+---+---+---+---+      Anwendungsversion:     4.5               //
//                                             Boardtypennummer:      6789              //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
static void
OnCallTypePlate(const Cy4Service* service, Cy4Message* message)
{
    static const UInt08  typePlate[][7] PROGMEM = {  { 'C','y','4','C','o','d','e' },
                                                     CY4TYPEPLATE(DEV_APPL_TYPE,
                                                                  DEV_APPL_VERS,
                                                                  DEV_BOARD_TYPE)
                                                  };

    UInt08 j = (UInt08)(CY4MESSAGE_Location(message) - CY4SERVICE_Base(service));
    for (UInt08 i=0; i<CY4MESSAGE_Count(message); i++)
    {   CY4MESSAGE_Data(message)[i] = pgm_read_byte(&typePlate[1][j]);
        j++;
    }
}

//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Pflichtdienst: EnterApplication (write only)                                        //
//                                                                                      //
//  Beschreibung: Startet die Anwendung neu.                                            //
//                                                                                      //
//              0                                                                       //
//            +---+                                                                     //
//  Belegung: |EAP|                                                                     //
//            +---+                                                                     //
//                                                                                      //
//  [EAP] Auslöser zum (Neu-)Start der Applikation (Reset): 94 (5Eh, '^').              //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
static void
OnEnterApplication(const Cy4Service* service, Cy4Message* message)
{
    (void)service;
    if (CY4MESSAGE_ConstData(message)[0] == 0x5E) BOARD_Reset();
    else                                          CY4MESSAGE_Invalidate(message);
}

//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Pflichtdienst: EnterProgramLoader (write only)                                      //
//                                                                                      //
//  Beschreibung: Startet das System neu und kehrt im Programmlader-Modul zurück.       //
//                                                                                      //
//              0                                                                       //
//            +---+                                                                     //
//  Belegung: |EPL|                                                                     //
//            +---+                                                                     //
//                                                                                      //
//  [EPL] Auslöser zum Eintritt in den Programmlader: 94 (5Eh, '^').                    //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
static void
OnEnterProgramLoader(const Cy4Service* service, Cy4Message* message)
{
    (void)service;
    if (CY4MESSAGE_ConstData(message)[0] == 0x5E) BOARD_EnterPgmLdr();
    else                                          CY4MESSAGE_Invalidate(message);
}

//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Pflichtdienst: ChangeCy4Address (write only)                                        //
//                                                                                      //
//  Beschreibung: Wechsel der Geräteadresse.                                            //
//                                                                                      //
//              0   1                                                                   //
//            +---+---+                                                                 //
//  Belegung: |NET|DEV|                                                                 //
//            +---+---+                                                                 //
//                                                                                      //
//  [NET] Neue Netzadresse:   1 .. 255                                                  //
//  [DEV] Neue Geräteadresse: 1 .. 255                                                  //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
static void
OnChangeCy4Address(const Cy4Service* service, Cy4Message* message)
{
    UInt08  data[2];
    {   Cy4Address address = CY4NET_Address();
        data[0] = CY4ADDRESS_Net(&address);
        data[1] = CY4ADDRESS_Dev(&address);
    }
    CY4SERVICE_TransferPayload(service, message, data);
    {   Cy4Address address = CY4ADDRESS(data[0], data[1]);
        if (CY4ADDRESS_IsBroadcast(&address) || !CY4ADDRESS_IsValid(&address))
        {   CY4MESSAGE_Invalidate(message);
        }
        else
        {   CY4NET_SetAddress(address);
            STORING_SaveCy4SelfAdr(address);
        }
    }
}

//////////////////////////////////////////////////////////////////////////////////////////
//  Prozessaufrufe  //////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
void
APPLICATION_Startup(void)
{
    static const Cy4ServiceItem  servicesTable[] PROGMEM =
    {   //|                   service                    |           handler           |
        //+----------------------------------------------+-----------------------------+
        { CY4SERVICE(0x130000,  3, CY4SERVICE_READ_ONLY) , OnContactState              },
        { CY4SERVICE(0x130010, 18, CY4SERVICE_READ_WRITE), OnContactStateSignal        },
        { CY4SERVICE(0x130100, 61, CY4SERVICE_READ_WRITE), OnContactEventConfiguration },
        { CY4SERVICE(0x130180, 61, CY4SERVICE_READ_WRITE), OnContactEventConfiguration },
        { CY4SERVICE(0x130200, 61, CY4SERVICE_READ_WRITE), OnContactEventConfiguration },
        { CY4SERVICE(0x130280, 61, CY4SERVICE_READ_WRITE), OnContactEventConfiguration },
        { CY4SERVICE(0x130300, 61, CY4SERVICE_READ_WRITE), OnContactEventConfiguration },
        { CY4SERVICE(0x130380, 61, CY4SERVICE_READ_WRITE), OnContactEventConfiguration },		
		{ CY4SERVICE(0x140000,  6, CY4SERVICE_WRITE_ONLY), OnContactEventTest          },	
        { CY4SERVICE(0xFFF020,  2, CY4SERVICE_READ_WRITE), OnBusActivityOnLed          },
        { CY4SERVICE(0xFFFF00, 61, CY4SERVICE_READ_WRITE), OnDeviceLabel               },
        { CY4SERVICE(0xFFFF40,  1, CY4SERVICE_WRITE_ONLY), OnFactorySettings           },
        { CY4SERVICE(0xFFFFF0,  7, CY4SERVICE_READ_ONLY) , OnCallTypePlate             },
        { CY4SERVICE(0xFFFFF8,  1, CY4SERVICE_WRITE_ONLY), OnEnterApplication          },
        { CY4SERVICE(0xFFFFFA,  1, CY4SERVICE_WRITE_ONLY), OnEnterProgramLoader        },
        { CY4SERVICE(0xFFFFFC,  2, CY4SERVICE_WRITE_ONLY), OnChangeCy4Address          }
    };
    CY4NET_AssignServices(servicesTable);
    Cy4Address address = STORING_LoadCy4SelfAdr();
    CY4NET_SetAddress(address);
}

//////////////////////////////////////////////////////////////////////////////////////////
void
APPLICATION_Proceed(void)
{
    ProceedContacting();
    ProceedSignaling();
    ProceedPushMessaging();
    ProceedTextFeeds();
}

//////////////////////////////////////////////////////////////////////////////////////////
void
APPLICATION_Shutdown(void)
{
}

//////////////////////////////////////////////////////////////////////////////////////////
